﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                   
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using Newtonsoft.Json;
using System.Runtime.Serialization.Formatters;

namespace PuppetMaster
{
    //
    // ─── MOVESET CLASS ────────────────────────────────────────────────────────────────
    //
    public class MoveSet
    {
        public Puppet Puppet;
        public int poseID        = -1;
        public string MovesList  = "";
        public bool doReset      = false;
        public bool autoStart    = false;
        public bool comboMove    = false;

        public readonly RagdollPose Ragdoll;

        public Dictionary<string, LimbMove> moves = new Dictionary<string, LimbMove>();

        public string MoveHash => "PM" + MovesList.GetHashCode();

        public struct LimbMove
        {
            public string limbName;
            public float angle;
            public float force;
       }

        //
        // ─── NEW MOVESET ────────────────────────────────────────────────────────────────
        //
        public MoveSet(string movesList, bool autoStart=true, RagdollPose presetPose=null)
        {
            MovesList = movesList;
            Puppet    = PuppetMaster.Puppet;
            Ragdoll   = new RagdollPose()
            {
                Name                     = MoveHash,
                Rigidity                 = 4,
                ShouldStandUpright       = true,
                DragInfluence            = 1,
                UprightForceMultiplier   = 1,
                ShouldStumble            = true,
                State                    = 0,
                AnimationSpeedMultiplier = 1,
                Angles                   = new List<RagdollPose.LimbPose>(),
            };

            if (presetPose != null)
            {
                Ragdoll      = presetPose.ShallowClone();
                Ragdoll.Name = MoveHash;
                comboMove    = true;
            }

            if (autoStart) Import();
        }


        public string[] GetPoseData(string movesList)
        {
            if (movesList.Contains(",")) return MovesList.Trim().Split(',');

            if (MoveData.ContainsKey(movesList)) return MoveData[movesList].Split(',');

            return null;

        }

        //
        // ─── IMPORT MOVESET ────────────────────────────────────────────────────────────────
        //
        public void Import()
        {
            string[] mvItem = GetPoseData(MovesList);

            for (int i = mvItem.Length; --i >= 0;)
            {
                string[] parts  = mvItem[i].Trim().Split(':');
                if (parts.Length != 3) continue;
                
                //  Skip commented out limbs
                if ("#/".Contains("" + parts[0][0])) continue;

                LimbMove move   = new LimbMove()
                {
                    limbName    = parts[0],
                };

                if (float.TryParse(parts[1], out float angle)) move.angle = angle;
                if (float.TryParse(parts[2], out float force)) move.force = force;

                moves.Add(parts[0], move);
            }

            RagdollPose.LimbPose limbPose;

            bool includeMove;

            foreach (LimbBehaviour limb in Puppet.LB.Values)
            {
                includeMove = false;

                if (limb.HasJoint)
                {
                    if (moves.ContainsKey(limb.name))
                    {
                        includeMove = true;
                    }
                }

                if (includeMove)
                {
                    limbPose = new RagdollPose.LimbPose(limb, moves[limb.name].angle)
                    {
                        Name                 = MoveHash,
                        Limb                 = limb,
                        Animated             = false,
                        PoseRigidityModifier = 0f,
                        StartAngle           = 0f,
                        EndAngle             = 0f,
                        AnimationDuration    = 0f,
                        RandomInfluence      = 0f,
                        RandomSpeed          = 0f,
                        TimeOffset           = 0f,
                        AnimationCurve       = new AnimationCurve(),
                    };
                }
                else if(!comboMove)
                {
                    limbPose = new RagdollPose.LimbPose(limb,0f)
                    {
                        Name                 = MoveHash,
                        Limb                 = limb,
                        Animated             = false,
                        PoseRigidityModifier = 0f,
                        Angle                = 0f,
                        StartAngle           = 0f,
                        EndAngle             = 0f,
                        AnimationDuration    = 0f,
                        RandomInfluence      = 0f,
                        RandomSpeed          = 0f,
                        TimeOffset           = 0f,
                        AnimationCurve       = new AnimationCurve(),
                    };
                } else continue;

                if (comboMove)
                {
                    for (int i = Ragdoll.Angles.Count; --i >= 0;)
                    {
                        if (Ragdoll.Angles[i].Limb == limb)
                        {
                            Ragdoll.Angles[i] = limbPose;
                        }
                    }
                } else 
                {
                    Ragdoll.Angles.Add(limbPose);
                }
            }

            Ragdoll.ConstructDictionary();

            for (int pnum = Puppet.PBO.Poses.Count; --pnum >= 0;)
            {
                if (Puppet.PBO.Poses[pnum].Name == MoveHash)
                {
                    poseID = pnum;

                    Puppet.PBO.Poses[pnum] = Ragdoll;

                    return;
                }
            }

            Puppet.PBO.Poses.Add(Ragdoll);
            
            poseID = Puppet.PBO.Poses.Count -1;

        }


        public bool RunMove()
        {
            Puppet.PBO.OverridePoseIndex = poseID;
            return true;
        }


        //
        // ─── COMBINE MOVESET ────────────────────────────────────────────────────────────────
        //
        public void CombineMove()
        {
            //  Only pose applicable limbs (continue previous actions)
            foreach (KeyValuePair<string, LimbMove> pair in moves)
            {
                Puppet.LB[pair.Key].InfluenceMotorSpeed(Mathf.DeltaAngle(
                    Puppet.LB[pair.Key].Joint.jointAngle, pair.Value.angle * Puppet.PBO.AngleOffset) * pair.Value.force);

            }

        }

        public void ClearMove()
        {
            foreach (string limbName in moves.Keys)
            {
                if (Puppet.LB?[limbName])
                {
                    Puppet.LB[limbName].Health                = Puppet.LB[limbName].InitialHealth;
                    Puppet.LB[limbName].Numbness              = 0.0f;
                    Puppet.LB[limbName].HealBone();

                    Puppet.LB[limbName].CirculationBehaviour.BleedingRate  *= 0.5f;
                    Puppet.LB[limbName].PhysicalBehaviour.BurnProgress     *= 0.5f;
                    Puppet.LB[limbName].SkinMaterialHandler.AcidProgress   *= 0.5f;
                    Puppet.LB[limbName].SkinMaterialHandler.RottenProgress *= 0.5f;
                }
            }

            if (!Puppet.IsCrouching) Puppet.PBO.OverridePoseIndex = -1;
        }

        //
        // ─── POSE DATA ────────────────────────────────────────────────────────────────
        //
        private static readonly Dictionary<string, string> MoveData = new Dictionary<string, string>()
        {
            { "throw_1",
              @"UpperArmFront:53.87996:30,
                UpperArm:0.1090901:30,
                LowerLegFront:-0.01318056:30,
                LowerArmFront:-36.86349:30,
                LowerArm:-0.4033567:30,
                UpperLegFront:3.002661:30,
                UpperLeg:-0.003749454:30,
                LowerBody:-0.05592484:30,
                LowerLeg:-0.007670409:30"
            },
            { "throw_2",
              @"LowerArm:0.1230614:30,
                UpperArm:0.117004:30,
                LowerLeg:0.001100474:30,
                LowerLegFront:0.001414623:30,
                UpperArmFront:174.287:30,
                LowerArmFront:86.93764:30,
                UpperLeg:0.007332622:30,
                UpperLegFront:0.00374265:30"
            },
            { "crouch",
              @"Head:-0.0002471675:30,
                UpperArm:0.1:30,
                #UpperArmFront:0.1:30,
                UpperLegFront:-75.54669:30,
                UpperLeg:-0.03494752:30,
                LowerLeg:106.0683:30,
                LowerLegFront:78.07973:30,
                Foot:-24.21083:30,
                FootFront:0.4032795:30"
            },
            { "dive",
              @"UpperArm:-136.7963:20,
                LowerBody:-11.74754:30,
                UpperArmFront:-136.7963:20,
                UpperBody:-8.516563:30,
                LowerArm:-34.3288:30,
                UpperLeg:-74.00098:30,
                UpperLegFront:-67.9553:30,
                LowerArmFront:-33.82881:30,
                LowerLegFront:63.65865:30,
                LowerLeg:60.26822:30"
            },
            { "backflip_1",
              @"Foot:-34.77201:30,
                FootFront:-27.44459:30,
                LowerArm:-30.00827:30,
                LowerArmFront:-38.24268:30,
                LowerLeg:95.33749:30,
                LowerLegFront:100.3498:30,
                UpperArm:-157.7551:30,
                UpperArmFront:-149.6151:30,
                UpperLeg:-28.73837:30,
                UpperLegFront:-27.9621:30"
            },
            { "backflip_2",
              @"Foot:51.61726:30,
                FootFront:50.23609:30,
                Head:0.8721468:30,
                LowerArm:-30.56421:30,
                LowerArmFront:-22.44059:30,
                LowerBody:-1.73008:30,
                LowerLeg:39.59584:30,
                LowerLegFront:52.36701:30,
                MiddleBody:0:30,
                UpperArm:155.1894:30,
                UpperArmFront:-196.8672:30,
                UpperBody:-0.8053715:30,
                UpperLeg:0.4376444:30,
                UpperLegFront:-0.03437976:30"
            },
            { "jump",
              @"LowerArm:-95.09151:30,
                LowerArmFront:-85.78412:30,
                UpperArm:-376.2525:30,
                UpperArmFront:-369.4041:30,
                UpperLegFront:-111.5341:30,
                UpperLeg:-112.034:30,
                LowerLeg:133.9088:30,
                LowerLegFront:134.5014:30"
            },
            { "prone",
              @"Head:-6.500257:30, 
                LowerBody:17.68937:30,
                UpperArmFront:-18.54307:30,
                UpperArm:-16.35537:30,
                LowerLegFront:9.899184:30,
                LowerLeg:11.33566:30,
                Foot:0.02769642:30, 
                FootFront:0.04737419:30,
                UpperLegFront:16.45826:30,
                UpperLeg:15.87339:30,
                LowerArmFront:-103.8956:30,
                UpperBody:-17.44694:30,
                LowerArm:-102.0114:30"
            },
        };
    }
}
